package dev.fitko.fitconnect.client.factory;

import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.core.SubmissionSender;
import dev.fitko.fitconnect.core.SubmissionSubscriber;
import dev.fitko.fitconnect.core.auth.DefaultOAuthService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.destination.DestinationRestService;
import dev.fitko.fitconnect.core.events.EventLogApiService;
import dev.fitko.fitconnect.core.events.SecurityEventTokenService;
import dev.fitko.fitconnect.core.http.ProxyConfig;
import dev.fitko.fitconnect.core.submission.SubmissionApiService;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.client.RestTemplate;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.stream.Collectors;

/**
 * Factory that constructs clients for {@link Sender} and {@link Subscriber}.
 */
public final class ClientFactory {

    private static final Logger LOGGER = LoggerFactory.getLogger(ClientFactory.class);

    private static final String CONFIG_DEFAULT_LOCATION = "config.yml";
    private static final String CUSTOM_CONFIG_KEY = "FIT_CONNECT_CONFIG";

    private static final String SENDER_BANNER = getSplashScreenResource("/sender-banner.txt");
    private static final String SUBSCRIBER_BANNER = getSplashScreenResource("/subscriber-banner.txt");

    private ClientFactory() {
    }

    /**
     * Create a new {@link SenderClient} that is automatically configured via the config.yml file.
     *
     * @return the sender client
     */
    public static SenderClient.Start senderClient() {
        final ApplicationConfig config = loadConfig();
        return senderClient(config);
    }

    /**
     * Create a new {@link SenderClient} that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the sender client
     */
    public static SenderClient.Start senderClient(final ApplicationConfig config) {
        LOGGER.info(SENDER_BANNER);
        LOGGER.info("Initializing sender client ...");
        final Sender sender = getSender(config);

        LOGGER.info("Reading metadata schema from {} ", config.getMetadataSchemaPath());
        final String metadataSchema = readPath(config.getMetadataSchemaPath());

        return SenderClient.build(sender, metadataSchema, config.getCurrentEnvironment());
    }


    /**
     * Create a new {@link SubscriberClient} that is automatically configured via config.yml file.
     *
     * @return the subscriber client
     */
    public static SubscriberClient.RequestSubmission subscriberClient() {
        final ApplicationConfig config = loadConfig();
        return subscriberClient(config);
    }

    /**
     * Create a new {@link SubscriberClient} that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the subscriber client
     */
    public static SubscriberClient.RequestSubmission subscriberClient(final ApplicationConfig config) {
        LOGGER.info(SUBSCRIBER_BANNER);
        LOGGER.info("Initializing subscriber client ...");
        final Subscriber subscriber = getSubscriber(config);
        final SubscriberConfig subscriberConfig = config.getSubscriberConfig();

        LOGGER.info("Reading private key from {} ", subscriberConfig.getPrivateDecryptionKeyPath());
        final String privateKey = readPath(subscriberConfig.getPrivateDecryptionKeyPath());

        LOGGER.info("Reading metadata schema from {} ", config.getMetadataSchemaPath());
        final String metadataSchema = readPath(config.getMetadataSchemaPath());

        return SubscriberClient.builder(subscriber, privateKey, metadataSchema);
    }

    private static Subscriber getSubscriber(final ApplicationConfig config) {
        final CryptoService cryptoService = getCryptoService();
        final ValidationService validator = getValidatorService(config);
        final RestTemplate restTemplate = getRestTemplate(config);
        final OAuthService authService = getSubscriberConfiguredAuthService(config, restTemplate);
        final EventLogService eventLogService = getEventLogService(config, restTemplate, authService);
        final SubmissionService submissionService = getSubmissionService(config, restTemplate, authService);
        final SecurityEventService setService = getSecurityEventTokenService(config);
        return new SubmissionSubscriber(submissionService, eventLogService, cryptoService, validator, setService);
    }

    private static Sender getSender(final ApplicationConfig config) {
        final CryptoService cryptoService = getCryptoService();
        final RestTemplate restTemplate = getRestTemplate(config);
        final ValidationService validator = getValidatorService(config);
        final OAuthService authService = getSenderConfiguredAuthService(config, restTemplate);
        final EventLogService eventLogService = getEventLogService(config, restTemplate, authService);
        final SubmissionService submissionService = getSubmissionService(config, restTemplate, authService);
        final DestinationService destinationService = getDestinationApiService(config, restTemplate, authService);
        final SecurityEventService setService = getSecurityEventTokenService(config);
        return new SubmissionSender(destinationService, submissionService, eventLogService, cryptoService, validator, setService);
    }

    private static OAuthService getSenderConfiguredAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        final String clientId = config.getSenderConfig().getClientId();
        final String clientSecret = config.getSenderConfig().getClientSecret();
        return new DefaultOAuthService(restTemplate, clientId, clientSecret, config.getOAuthTokenEndpoint());
    }

    private static OAuthService getSubscriberConfiguredAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        final String clientId = config.getSubscriberConfig().getClientId();
        final String clientSecret = config.getSubscriberConfig().getClientSecret();
        return new DefaultOAuthService(restTemplate, clientId, clientSecret, config.getOAuthTokenEndpoint());
    }

    private static SubmissionService getSubmissionService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new SubmissionApiService(authService, restTemplate, config);
    }

    private static EventLogService getEventLogService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new EventLogApiService(authService, restTemplate, config);
    }

    private static DestinationRestService getDestinationApiService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new DestinationRestService(authService, restTemplate, config);
    }

    private static DefaultValidationService getValidatorService(final ApplicationConfig config) {
        final MessageDigestService messageDigestService = getMessageDigestService();
        return new DefaultValidationService(messageDigestService, config);
    }

    private static CryptoService getCryptoService() {
        final MessageDigestService messageDigestService = getMessageDigestService();
        return new JWECryptoService(messageDigestService);
    }

    private static RestTemplate getRestTemplate(final ApplicationConfig config) {
        final ProxyConfig proxyConfig = new ProxyConfig(config.getHttpProxyHost(), config.getHttpProxyPort());
        return proxyConfig.proxyRestTemplate();
    }

    private static MessageDigestService getMessageDigestService() {
        return new HashService();
    }

    private static SecurityEventTokenService getSecurityEventTokenService(final ApplicationConfig config) {
        LOGGER.info("Reading private signing key from {} ", config.getPrivateSigningKeyPath());
        final String signingKey = readPath(config.getPrivateSigningKeyPath());
        return new SecurityEventTokenService(signingKey);
    }

    private static String readPath(final String pathName) {
        try {
            return Files.readString(Path.of(pathName));
        } catch (final IOException e) {
            LOGGER.error("Path {} could not be loaded", pathName);
        }
        return null;
    }

    private static String readPathFromEnvironment(final String variableName) {
        try {
            return System.getenv(variableName);
        } catch (final NullPointerException e) {
            LOGGER.error("Environment variable {} could not be loaded", variableName);
            return null;
        }
    }

    private static ApplicationConfig loadConfig() {
        final String customConfigLocation = readPathFromEnvironment(CUSTOM_CONFIG_KEY);
        if (customConfigLocation == null) {
            return loadConfigFromPath(CONFIG_DEFAULT_LOCATION);
        }
        return loadConfigFromPath(customConfigLocation);
    }

    private static ApplicationConfig loadConfigFromPath(final String configPath) {
        try {
            final Path pathToConfig = Path.of(configPath);
            final ApplicationConfig applicationConfig = ApplicationConfigLoader.loadConfig(pathToConfig);
            LOGGER.info("Using sdk environment config {} with {}", applicationConfig.getCurrentEnvironment(), applicationConfig);
            return applicationConfig;
        } catch (final InitializationException e) {
            LOGGER.error("Config could not be loaded, please provide a 'config.yml' in the root path or in the ENV-var 'FIT_CONNECT_CONFIG'");
            throw e;
        }
    }

    private static String getSplashScreenResource(final String bannerName) {
        final InputStream is = ClientFactory.class.getResourceAsStream(bannerName);
        return new BufferedReader(new InputStreamReader(is)).lines().collect(Collectors.joining("\n"));
    }
}




